import pytest


@pytest.mark.xfail
def test_validate_feature_api():
    raise NotImplementedError


@pytest.mark.xfail
def test_validate_feature_acceptance():
    raise NotImplementedError


@pytest.mark.xfail
def test_discover():
    pass
